import numpy as np
from src.plot import plot_xy
import time
from src.TiseSolverFgh1D import TiseSolverFgh1D
from src.TdseSolverScipy1D import TdseSolverScipy1D
import os
import imageio.v2 as imageio  # Explicitly specify v2


if __name__ == "__main__":
    # - Input parameters
    # -- Position
    xa = -10.0  # Position x = [xa, xb]
    xb = 10.0
    nx = 2**6  # Number of x points

    # -- Time
    ta = 0.0  # Start time
    tb = 2 * np.pi * 2  # End time
    nt = 16 * 2 + 1  # Number of output time

    # -- Tolerance for time propagation
    rtol = 1e-6  # Relative tolerance
    atol = 1e-6  # Absolute tolerance
    method = "RK45"  # Integration method

    # -- Output folder
    output_folder_path = "./out/"

    # =====================
    #         TISE
    # =====================

    # -- Solve TISE using Fourier-Grid Hamiltonian method
    solver = TiseSolverFgh1D(xa, xb, nx)
    x, Vx = solver.abscissa_potential()
    start_time = time.time()  # Record the start time
    eigenvalues, eigenfunctions = solver.solve()
    end_time = time.time()  # Record the end time
    print(f"  Elapsed time TISE: {end_time - start_time:.6f} seconds")

    # -- Output TISE results
    x_dict = {}
    y_dict = {}
    x_dict["V(x)"] = np.array(x)
    y_dict["V(x)"] = np.array(Vx)
    for n in range(5):
        x_dict[f"Idx {n:>2d}"] = np.array(x)
        y_dict[f"Idx {n:>2d}"] = eigenfunctions[:, n].real + eigenvalues[n]
    output_image_name = os.path.join(output_folder_path, "eigenfunctions.png")
    plot_xy(
        x_dict,
        y_dict,
        show_gui=False,
        output_image_name=output_image_name,
    )

    # - Eigenvalues output
    output_csv_name = os.path.join(output_folder_path, "tise_eigenvalues.csv")
    with open(output_csv_name, "w") as f:
        for idx, value in enumerate(eigenvalues):
            f.write(f"{idx},{value}\n")
    # =====================
    #         TDSE
    # =====================

    # -- Output time (not affect to the integration tolerance)
    t = np.linspace(ta, tb, nt, endpoint=True, dtype=np.float64)

    # -- Solve TDSE with below methods
    #      Time integral       : scipy.integrate.solve_ivp
    #      Position derivative : compute the derivative through FFT
    solver = TdseSolverScipy1D(xa, xb, nx, d_method="fft")
    x = solver.abscissa()
    # -- Initial state (FGH grid and TDSE grid is coincide)
    psi_ini = ((eigenfunctions[:, 0] + eigenfunctions[:, 1]) / np.sqrt(2.0)).astype(np.complex128)
    start_time = time.time()  # Record the start time
    sol = solver.solve(t, psi_ini, rtol=rtol, atol=atol, method=method)
    end_time = time.time()  # Record the end time
    print(f"  Elapsed time TDSE: {end_time - start_time:.6f} seconds")

    # -- TDSE Analysis
    # --- projection
    projection = solver.project_to_eigenstates(eigenfunctions)

    # -- Output TDSE results
    # --- Make output folder
    os.makedirs(output_folder_path, exist_ok=True)

    # --- Output projection results
    x_dict = {}
    y_dict = {}
    for i in range(10):
        x_dict[f"State {i}"] = np.array(t)
        y_dict[f"State {i}"] = np.abs(projection[:, i])
    output_image_name = os.path.join(output_folder_path, "projection.png")
    plot_xy(x_dict, y_dict, y_range=(0, 1), show_gui=False, output_image_name=output_image_name)

    # --- Output wave function
    start_time = time.time()  # Record the start time
    frames = []
    for i, ti in enumerate(t):
        x_dict = {}
        y_dict = {}
        x_dict["Re$(\\psi)$"] = np.array(x)
        y_dict["Re$(\\psi)$"] = sol[i, :].real
        x_dict["Im$(\\psi)$"] = np.array(x)
        y_dict["Im$(\\psi)$"] = sol[i, :].imag

        output_image_name = os.path.join(output_folder_path, f"frame_{i:03d}.png")
        data_optional = {
            "Re$(\\psi)$": {"fmt": "o-", "lw": 2, "lc": "C0"},
            "Im$(\\psi)$": {"fmt": "o-", "lw": 2, "lc": "C1"},
        }
        plot_xy(
            x_dict,
            y_dict,
            y_range=(-2.0, 2.0),
            data_optional=data_optional,
            title_name=f"t={ti:.3f}",
            figure_size=(10, 5),
            show_gui=False,
            image_dpi=100,
            output_image_name=output_image_name,
        )
        frames.append(imageio.imread(output_image_name))

    # Save as GIF
    output_gif = os.path.join(output_folder_path, "wavefunction.gif")
    imageio.mimsave(output_gif, frames, fps=5, loop=0)
    end_time = time.time()  # Record the end time
    print(f"  Elapsed time  GIF: {end_time - start_time:.6f} seconds")
