import scipy as sp
import numpy as np
from findiff import Diff
import pyfftw


class TdseSolverScipy1D:
    def __init__(
        self,
        xa: float,
        xb: float,
        nx: int,
        d_method: str = "fft",
    ):
        self.xa: float = xa
        self.xb: float = xb
        self.nx: int = nx
        x = np.linspace(xa, xb, nx, endpoint=False, dtype=np.float64)
        self.x: np.ndarray = x

        dx = x[1] - x[0]
        self.d_method = d_method
        if self.d_method == "fft":
            self.k = np.fft.fftfreq(self.nx, dx)
            planner_effort = "FFTW_MEASURE"
            self.fft_objects = pyfftw.builders.fft(np.empty(self.nx, dtype=np.complex128), planner_effort=planner_effort)  # FFT実行用オブジェクト生成
            self.ifft_objects = pyfftw.builders.ifft(np.empty(self.nx, dtype=np.complex128), planner_effort=planner_effort)  # IFFT実行用オブジェクト生成
            self.ck1 = 1.0j * 2.0 * np.pi * self.k
            self.ck2 = -((2.0 * np.pi * self.k) ** 2)
        elif self.d_method == "findiff":
            self.d_dx = Diff(0, dx, acc=4, periodic=True)  # 4th order (5 points) with periodic condition
            self.d2_dx2 = self.d_dx**2

    def abscissa(self):
        return self.x

    # TDSE Coefficient applied to the 2nd-order differential operator
    def C2(self, t: float, psi: np.ndarray):
        c = -1.0 / 2.0
        return c

    # TDSE Coefficient applied to the 1st-order differential operator
    def C1(self, t: float, psi: np.ndarray):
        c = 0.0
        return c

    # TDSE Coefficient applied to the 0th-order differential operator
    def C0(self, t: float, psi: np.ndarray):
        c = 0.5 * self.x**2
        return c

    def solve(
        self,
        t: np.ndarray,
        psi0: np.ndarray,
        rtol: float = 1e-9,
        atol: float = 1e-9,
        method: str = "RK45",
    ) -> np.ndarray:
        if self.d_method == "fft":
            sol = sp.integrate.solve_ivp(self.Hpsi_fft, [t[0], t[-1]], psi0, method=method, t_eval=t, rtol=rtol, atol=atol)
        elif self.d_method == "findiff":
            sol = sp.integrate.solve_ivp(self.Hpsi_findiff, [t[0], t[-1]], psi0, method=method, t_eval=t, rtol=rtol, atol=atol)

        self.sol = sol.y.T

        return self.sol

    def solve_section(
        self,
        t: np.ndarray,
        psi0: np.ndarray,
        t_section: float = 1.0,
        rtol: float = 1e-9,
        atol: float = 1e-9,
        method: str = "RK45",
    ) -> np.ndarray:
        ta = t[0]
        tb = t[-1]
        # t_step_rough = 2.0
        n_section = np.floor((tb - ta) / t_section).astype(int) + 1
        t_sections = np.linspace(ta, tb, n_section)

        solution = np.zeros((len(t), len(psi0)), dtype=psi0.dtype)
        idx = 0
        y0 = psi0.copy()
        for k in range(n_section - 1):
            t_start = t_sections[k]
            t_end = t_sections[k + 1]

            # # `t` のサブセットを取得
            if np.isclose(t_end, tb, rtol=rtol, atol=atol):
                t_subset = t[(t >= t_start) & (t <= t_end)]  # 最後の区間であれば
            else:
                t_subset = t[(t >= t_start) & (t < t_end)]

            add_t_start = not (len(t_subset) > 0 and np.isclose(t_subset[0], t_start, rtol=rtol, atol=atol))
            add_t_end = not (len(t_subset) > 0 and np.isclose(t_subset[-1], t_end, rtol=rtol, atol=atol))
            if add_t_start:
                t_subset = np.insert(t_subset, 0, t_start)
            if add_t_end:
                t_subset = np.append(t_subset, t_end)

            if self.d_method == "fft":
                sol = sp.integrate.solve_ivp(self.Hpsi_fft, (t_subset[0], t_subset[-1]), y0, method=method, t_eval=t_subset, rtol=rtol, atol=atol)
            elif self.d_method == "findiff":
                sol = sp.integrate.solve_ivp(self.Hpsi_findiff, (t_subset[0], t_subset[-1]), y0, method=method, t_eval=t_subset, rtol=rtol, atol=atol)

            # t以外に付け加えた要素を除外して解に代入する
            start_idx = 1 if add_t_start else 0
            end_idx = len(t_subset) - 1 if add_t_end else len(t_subset)

            solution[idx : idx + (end_idx - start_idx)] = sol.y[:, start_idx:end_idx].T
            idx += end_idx - start_idx

            y0 = sol.y[:, -1]  # 次の区間の初期条件として使用

        # solution = solution.T
        self.sol = solution
        return self.sol

    def Hpsi_findiff(
        self,
        t: float,
        psi_x: np.ndarray,
    ):
        psi_x1 = self.d_dx(psi_x)  # \frac{d}{dx}\psi(x)
        psi_x2 = self.d2_dx2(psi_x)  # \frac{d^2}{dx^2}\psi(x)
        rhs = -1.0j * (self.C2(t, psi_x) * psi_x2 + self.C1(t, psi_x) * psi_x1 + self.C0(t, psi_x) * psi_x)

        return rhs

    def Hpsi_fft(
        self,
        t: float,
        psi_x: np.ndarray,
    ):
        psi_k = self.fft_objects(psi_x)
        psi_x1 = self.ifft_objects(self.ck1 * psi_k)  # \frac{d}{dx}\psi(x)
        psi_x2 = self.ifft_objects(self.ck2 * psi_k)  # \frac{d^2}{dx^2}\psi(x)
        rhs = -1.0j * (self.C2(t, psi_x) * psi_x2 + self.C1(t, psi_x) * psi_x1 + self.C0(t, psi_x) * psi_x)

        return rhs

    def project_to_eigenstates(self, eigenfunctions: np.ndarray):
        # --- Projection to eigenstates (Quadrature for FGH is rectangle method)
        #     project_to_eigenstates[i:time idx, k:eigenstate idx] = sum_{n=0}^{nx-1} sol_fft[i, n] * eigenfunctions[n, k]
        projection = np.dot(self.sol, eigenfunctions) * ((self.xb - self.xa) / self.nx)
        return projection
